;; Generated package description from parser-generator.el  -*- no-byte-compile: t -*-
(define-package "parser-generator" "0.1.4.0.20220212.72656" "Parser Generator library" '((emacs "26")) :commit "14f21396e961607019af7f4c190a32a4f9b361fc" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-parser-generator")
